/*
 * Copyright (c) 2014-2015 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Nov 20 2014
 */
#include "acheader.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/atomic.h>
#include <util/twi.h>

#include "tk3-mikrokopter.h"

#if defined(__AVR_ATmega1284P__)
# define TK3_TWI_MASTER
#elif defined(__AVR_ATmega168A__)
# define TK3_TWI_SLAVE
#else
# error "unknown mcu"
#endif

#ifdef UBRR1
# define TK3_UART_N	2
#else
# define TK3_UART_N	1
#endif


/* --- local data ---------------------------------------------------------- */

struct tk3_iobuffer {
  volatile uint8_t recvbuf[64];
  volatile uint8_t recv_r, recv_w;

  uint8_t tsflag;
  volatile tk3_time recv_ts;

  volatile uint8_t sendbuf[64];
  volatile uint8_t send_r, send_w;
};

static struct tk3_iobuffer uart[TK3_UART_N];
static struct tk3_iobuffer twi;

#ifdef TK3_TWI_MASTER
static uint8_t twi_nmotors;
#endif
#ifdef TK3_TWI_SLAVE
static uint8_t twi_addr;
#endif

static uint8_t	tk3_sndbuffer_len(const struct tk3_iobuffer *io);
static uint8_t	tk3_rcvbuffer_len(const struct tk3_iobuffer *io);
static int	tk3_io_recv(struct tk3_iobuffer *io, struct tk3_iorecv *iorecv);
static void	tk3_encodechar(char c, struct tk3_iobuffer *io);
static void	tk3_putchar(char c, struct tk3_iobuffer *io);
static void	tk3_getchar(char c, struct tk3_iobuffer *io);


/* --- tk3_uart_init ------------------------------------------------------- */

int8_t
tk3_uart_init(uint8_t unit, uint32_t baud)
{
  uint16_t b = F_CPU / baud / 8;

  /* round to the nearest integer */
  if (F_CPU % (baud * 8) <= baud * 4)
    b--;

  /* helper macro */
#define TK3_UART_INIT(n)                                        \
  do {                                                          \
    /* set baudrate */                                          \
    UBRR ## n = b;                                              \
                                                                \
    /* double clock speed */                                    \
    UCSR ## n ## A = (1 << U2X ## n);                           \
    /* enable UARTn for transmission and reception */           \
    UCSR ## n ## B =                                            \
      (1 << RXCIE ## n) | (1 << RXEN ## n) | (1 << TXEN ## n);  \
    /* 8/N/1 */                                                 \
    UCSR ## n ## C =                                            \
      (1 << UCSZ ## n ## 1) | (1 << UCSZ ## n ## 0);            \
                                                                \
    /* initialize buffers */                                    \
    uart[n].recv_r = 0;                                         \
    uart[n].recv_w = 0;                                         \
    uart[n].send_r = 0;                                         \
    uart[n].send_w = 0;                                         \
  } while(0)

  /* initialize given unit */
  switch(unit) {
    case 0: TK3_UART_INIT(0); break;
#if TK3_UART_N > 1
    case 1: TK3_UART_INIT(1); break;
#endif
    default: return -1;
  }

  return 0;
}


/* --- tk3_twi_init -------------------------------------------------------- */

void
tk3_twi_init(uint8_t addr)
{
#ifdef TK3_TWI_MASTER
  /* 500kHz bus - max speed for 8MHz BLs */
  TWSR &= ~(( 1 << TWPS1) | (1 << TWPS0));
  TWBR = ((F_CPU / 500000) - 16) / (2 * 1);
#endif

  TWAR = (1 << TWGCE); /* own address is programmed when data is to be sent */
  TWCR = (1 << TWEA) | (1 << TWEN) | (1 << TWIE);

#ifdef TK3_TWI_MASTER
  /* start continuously polling for slave messages */
  TWCR |= (1 << TWINT) | (1 << TWSTA);
  twi_nmotors = addr;
#endif

#ifdef TK3_TWI_SLAVE
  twi_addr = addr;
#endif

  /* initialize buffers */
  twi.recv_r = 0;
  twi.recv_w = 0;
  twi.send_r = 0;
  twi.send_w = 0;
}


/* --- tk3_buffer_len ------------------------------------------------------ */

static inline uint8_t
tk3_sndbuffer_len(const struct tk3_iobuffer *io)
{
  uint8_t l;

  l = (io->send_w < io->send_r) ? sizeof(io->sendbuf) : 0;
  l += io->send_w;
  l -= io->send_r;
  return l * 100 / sizeof(io->sendbuf);
}

static inline uint8_t
tk3_rcvbuffer_len(const struct tk3_iobuffer *io)
{
  uint8_t l;

  l = (io->recv_w < io->recv_r) ? sizeof(io->recvbuf) : 0;
  l += io->recv_w;
  l -= io->recv_r;
  return l * 100 / sizeof(io->recvbuf);
}


/* --- tk3_log ------------------------------------------------------------- */

int
tk3_log(enum tk3_channel channel, const char *fmt, ...)
{
  struct tk3_iobuffer *io;
  va_list ap;
  char c;

  switch(channel) {
    case TK3_UART0: io = &uart[0]; break;
#if TK3_UART_N > 1
    case TK3_UART1: io = &uart[1]; break;
#endif
    case TK3_TWI: io = &twi; break;
    default: return 1;
  }

  va_start(ap, fmt);
  tk3_putchar('^', io);
  while((c = pgm_read_byte(fmt++))) {
    switch(c) {
      case '%': {
        switch(pgm_read_byte(fmt++)) {
          case '1':
            tk3_encodechar(va_arg(ap, int), io);
            break;

          case '2': {
            uint16_t x = va_arg(ap, uint16_t);
            tk3_encodechar((x >> 8) & 0xff, io);
            tk3_encodechar(x & 0xff, io);
            break;
          }

          case '4': {
            uint32_t x = va_arg(ap, uint32_t);
            tk3_encodechar((x >> 24) & 0xff, io);
            tk3_encodechar((x >> 16) & 0xff, io);
            tk3_encodechar((x >> 8) & 0xff, io);
            tk3_encodechar(x & 0xff, io);
            break;
          }
        }
        break;
      }

      default:
        tk3_encodechar(c, io);
    }
  }
  tk3_putchar('$', io);
  va_end(ap);
  return 0;
}


/* --- tk3_log_buffer ------------------------------------------------------ */

int
tk3_log_buffer(enum tk3_channel channel, const uint8_t *buffer, uint8_t len)
{
  struct tk3_iobuffer *io;

  switch(channel) {
    case TK3_UART0: io = &uart[0]; break;
#if TK3_UART_N > 1
    case TK3_UART1: io = &uart[1]; break;
#endif
    case TK3_TWI: io = &twi; break;
    default: return 1;
  }

  tk3_putchar('^', io);
  while(len--)
    tk3_encodechar(*buffer++, io);
  tk3_putchar('$', io);
  return 0;
}


/* --- tk3_io_recv --------------------------------------------------------- */

struct tk3_iorecv *
tk3_recv(void)
{
  static struct tk3_iorecv io_twi = { .state.channel = TK3_TWI };
  static struct tk3_iorecv io_uart0 = { .state.channel = TK3_UART0 };
#if TK3_UART_N > 1
  static struct tk3_iorecv io_uart1 = { .state.channel = TK3_UART1 };
#endif

  if (tk3_rcvbuffer_len(&twi) > 75) {
    if (tk3_comm_overflow) tk3_comm_overflow();
  }
  if (tk3_io_recv(&twi, &io_twi)) return &io_twi;

  if (tk3_rcvbuffer_len(&uart[0]) > 75) {
    if (tk3_comm_overflow) tk3_comm_overflow();
  }
  if (tk3_io_recv(&uart[0], &io_uart0)) return &io_uart0;

#if TK3_UART_N > 1
  if (tk3_rcvbuffer_len(&uart[1]) > 75) {
    if (tk3_comm_overflow) tk3_comm_overflow();
  }
  if (tk3_io_recv(&uart[1], &io_uart1)) return &io_uart1;
#endif

  if (tk3_comm_noverflow) tk3_comm_noverflow();
  return NULL;
}

static int
tk3_io_recv(struct tk3_iobuffer *io,  struct tk3_iorecv *iorecv)
{
  uint8_t r;
  char c;

  r = io->recv_r;
  while(r != io->recv_w) {
    c = io->recvbuf[r];
    r = (r + 1) % sizeof(io->recvbuf);
    io->recv_r = r;

    switch (c) {
      case '^':
        iorecv->state.start = 1;
        iorecv->state.escape = 0;
        iorecv->ts = io->recv_ts;
        iorecv->len = 0;
        break;

      case '$':
        if (!iorecv->state.start) break;
        iorecv->state.start = 0;
        return 1;

      case '!':
        iorecv->state.start = 0;
        break;

      case '\\':
        iorecv->state.escape = 1;
        break;

      default:
        if (!iorecv->state.start) break;
        if (iorecv->len >= sizeof(iorecv->data)) {
          iorecv->state.start = 0;
          break;
        }
        if (iorecv->state.escape) {
          iorecv->state.escape = 0;
          c = ~c;
        }

        iorecv->data[iorecv->len++] = c;
        break;
    }
  }

  return 0;
}


/* --- tk3_send ------------------------------------------------------------ */

void
tk3_send()
{
  uint8_t c[TK3_UART_N] = {}; /* silence wrong uninitialized warning */
  uint8_t r[TK3_UART_N];
  uint8_t f;

  /* twi */
  if (tk3_sndbuffer_len(&twi) > 75) {
    if (tk3_comm_overflow) tk3_comm_overflow();
  }
#ifdef TK3_TWI_SLAVE
  if (twi.send_r != twi.send_w) {
    /* listen to our address to reply to master polls */
    TWAR = (twi_addr << 1) | (1 << TWGCE);
  }
#endif

  /* uart (synchronous, with all units in parallel) */
  if (tk3_sndbuffer_len(&uart[0]) > 75) {
    if (tk3_comm_overflow) tk3_comm_overflow();
  }
  r[0] = uart[0].send_r;
#if TK3_UART_N > 1
  if (tk3_sndbuffer_len(&uart[1]) > 75) {
    if (tk3_comm_overflow) tk3_comm_overflow();
  }
  r[1] = uart[1].send_r;
#endif
  f = 0;
  do {
    /* helper macro */
#define TK3_UART_SEND(n)                                        \
    do {                                                        \
      switch (f & (1 << (n))) {                                 \
        case 1 << (n):                                          \
          if (!(UCSR ## n ## A & (1 << UDRE ## n))) break;      \
                                                                \
          UDR ## n = c[n];                                      \
          f &= ~(1 << (n));                                     \
          /* FALLTHROUGH */                                     \
        case 0:                                                 \
          if (r[n] != uart[n].send_w) {                         \
            c[n] = uart[n].sendbuf[r[n]];                       \
            r[n] = (r[n] + 1) % sizeof(uart[n].sendbuf);        \
            uart[n].send_r = r[n];                              \
            f |= 1 << (n);                                      \
          }                                                     \
          break;                                                \
      }                                                         \
    } while(0)

    /* scan all units */
    TK3_UART_SEND(0);
#if TK3_UART_N > 1
    TK3_UART_SEND(1);
#endif
  } while(f);
}


/* --- tk3_encodechar ------------------------------------------------------ */

static void
tk3_encodechar(char c, struct tk3_iobuffer *io)
{
  switch (c) {
    case '^': case '$':  case '!': case '\\':
      tk3_putchar('\\', io);
      c = ~c;
      break;
  }
  tk3_putchar(c, io);
}


/* --- tk3_putchar --------------------------------------------------------- */

static void
tk3_putchar(char c, struct tk3_iobuffer *io)
{
  uint8_t w, wp, r, rm;

  w = io->send_w;
  r = io->send_r;
  wp = (w + 1) % sizeof(io->sendbuf);
  if (wp == r) {
    /* discard oldest message in case of buffer overflow */
    do {
      rm = r;
      r = (r + 1) % sizeof(io->sendbuf);
    } while(io->sendbuf[rm] != '$' && r != wp);
    io->send_r = r;
  }
  io->sendbuf[w] = c;
  io->send_w = wp;
}


/* --- tk3_getchar --------------------------------------------------------- */

static inline void
tk3_getchar(char c, struct tk3_iobuffer *io)
{
  uint8_t w, wp, r, rm;

  if (io->tsflag) {
    if (c == 't') {
      /* get local timestamp of the last received timestamp message */
      io->recv_ts = tk3_clock_gettime();
    }
    io->tsflag = 0;
  } else if (c == '^') io->tsflag = 1;

  /* push byte */
  w = io->recv_w;
  r = io->recv_r;
  wp = (w + 1) % sizeof(io->recvbuf);
  if (wp == r) {
    /* discard oldest message in case of buffer overflow */
    do {
      rm = r;
      r = (r + 1) % sizeof(io->recvbuf);
    } while(io->recvbuf[rm] != '$' && r != wp);
    io->recv_r = r;
  }
  io->recvbuf[w] = c;
  io->recv_w = wp;
}


/* --- USARTn_RX_vect ------------------------------------------------------ */

#ifndef USART0_RX_vect
# ifdef USART_RX_vect
#   define USART0_RX_vect USART_RX_vect
# endif
#endif

#define USARTn_RX_vect(n)                                               \
  do {                                                                  \
    uint8_t status = UCSR ## n ## A;                                    \
    uint8_t c = UDR ## n;                                               \
                                                                        \
    if (status & ((1 << FE ## n) | (1 << UPE ## n) | (1 << DOR ## n)))  \
      c = '!';                                                          \
                                                                        \
    tk3_getchar(c, &uart[n]);                                           \
  } while(0)

ISR(USART0_RX_vect) { USARTn_RX_vect(0); }
#if TK3_UART_N > 1
ISR(USART1_RX_vect) { USARTn_RX_vect(1); }
#endif


/* --- TWI_vect ------------------------------------------------------------ */

#define TWCR_CLRINT ((1 << TWINT) | (1 << TWEN) | (1 << TWIE))

ISR(TWI_vect)
{
  uint8_t r;

  switch (TW_STATUS) {

#ifdef TK3_TWI_MASTER
    /* bus start */

    case TW_START:
    case TW_REP_START: {
      /* start condition transmitted */
      /* or repeated start condition transmitted */
      static uint8_t sladdr;

      if (twi.send_r != twi.send_w) {
        /* highest priority is to broadcast data to the BLs */
        TWDR = 0 | TW_WRITE; /* general call */
      } else {
        /* otherwise, do round robin polling of the BLs */
        TWDR = ((TK3TWI_BASEADDR + sladdr + 1) << 1) | TW_READ;
        sladdr = (sladdr + 1) % twi_nmotors;
      }
      TWCR = TWCR_CLRINT;
      break;
    }

      /* master transmitter mode */

    case TW_MT_SLA_NACK:
      /* SLA+W transmitted, NACK received */
      TWCR = TWCR_CLRINT | (1 << TWSTA); /* repeated start */
      break;

    case TW_MT_SLA_ACK:
    case TW_MT_DATA_ACK:
    case TW_MT_DATA_NACK:
      /* SLA+W transmitted, ACK received */
      /* data transmitted, ACK received */
      /* data transmitted, NACK received */
      r = twi.send_r;
      if (r != twi.send_w) {
        TWDR = twi.sendbuf[r];
        TWCR = TWCR_CLRINT;
        twi.send_r = (r + 1) % sizeof(twi.sendbuf);
      } else
        TWCR = TWCR_CLRINT | (1 << TWSTA); /* repeated start */
      break;


      /* master receiver mode */

    case TW_MR_SLA_NACK:
      /* SLA+R transmitted, NACK received */
      TWCR = TWCR_CLRINT | (1 << TWSTA); /* repeated start */
      break;

    case TW_MR_SLA_ACK:
      /* SLA+R transmitted, ACK received */
      TWCR = TWCR_CLRINT | (1 << TWEA); /* ack next byte */
      break;

    case TW_MR_DATA_ACK: {
      /* data received, ACK returned */

      /* the strategy is to nack as soon as an end-of-message marker is
       * seen, because there is only one reception buffer and incomplete
       * messages from different BLs would not interleave correctly.
       * This makes an overhead of one byte,but any other strategy
       * (e.g. sending the buffer length ahead) will also consume extra bytes.
       * The BL will send fake 0xff after the '$' (normally only once). */
      uint8_t c = TWDR;
      if (c == '$')
        TWCR = TWCR_CLRINT; /* nack next byte */
      else
        TWCR = TWCR_CLRINT | (1 << TWEA); /* ack next byte */

      tk3_getchar(c, &twi);
      break;
    }

    case TW_MR_DATA_NACK:
      /* data received, NACK returned */

      /* this extra byte is the fake 0xff from the slave, sent after the
       * end-of-message. It can be ignored. */
      TWCR = TWCR_CLRINT | (1 << TWSTA); /* repeated start */
      break;

#endif /* TK3_TWI_MASTER */

#ifdef TK3_TWI_SLAVE

      /* slave transmitter mode */

    case TW_ST_SLA_ACK:
    case TW_ST_DATA_ACK: {
      /* SLA+R received, ACK returned */
      /* data transmitted, ACK received */
      uint8_t c;

      /* send a fake '$' whenever the send buffer is empty. This should not
       * happen. */
      r = twi.send_r;
      if (r != twi.send_w) {
        c = twi.sendbuf[r];
        twi.send_r = (r + 1) % sizeof(twi.sendbuf);
      } else
        c = '$';

      TWDR = c;
      if (c == '$')
        TWCR = TWCR_CLRINT; /* stop sending after first message */
      else
        TWCR = TWCR_CLRINT | (1 << TWEA);
      break;
    }

    case TW_ST_DATA_NACK:
    case TW_ST_LAST_DATA:
      /* data transmitted, NACK received */
      /* last data byte transmitted, ACK received */

      if (twi.send_r == twi.send_w) {
        /* disconnect from the bus by resetting our address, so that the master
         * gets a nack immediately when polling us. It is still required to
         * listen to general calls, so TWEA cannot just be disabled */
        TWAR = 0 | (1 << TWGCE);
      }
      TWCR = TWCR_CLRINT | (1 << TWEA);
      break;


      /* slave receiver mode */

    case TW_SR_GCALL_ACK:
      /* general call received, ACK returned */
      TWCR = TWCR_CLRINT | (1 << TWEA);
      break;

    case TW_SR_GCALL_DATA_ACK:
    case TW_SR_GCALL_DATA_NACK:
      /* or general call data received, ACK returned */
      /* or general call data received, NACK returned */
      tk3_getchar(TWDR, &twi);
      TWCR = TWCR_CLRINT | (1 << TWEA);
      break;

    case TW_SR_STOP:
      /* stop or repeated start condition received while selected */
      TWCR = TWCR_CLRINT | (1 << TWEA);
      break;

#endif /* TK3_TWI_SLAVE */

      /* error conditions */

    case TW_NO_INFO:
      /* no state information available */
      break;

    case TW_BUS_ERROR:
      /* illegal start or stop condition */
      TWCR = TWCR_CLRINT | (1 << TWSTO);
      break;

    default:
      /* this must not happen by construction */
      abort();
  }
}
